#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import (
    absolute_import, unicode_literals, division, print_function
    )

# from functools import partial, lru_cache
from astropy import units as apu
import numpy as np

from . import cyprop
from .. import conversions as cnv
from .. import utils
# import ipdb


__all__ = [
    'PathProp',
    'loss_freespace', 'loss_troposcatter', 'loss_ducting',
    'loss_diffraction', 'loss_complete',
    'clutter_correction', 'clutter_imt',
    'height_profile_data', 'atten_map_fast',
    ]

# Note, we have to curry the quantities here, because Cython produces
# "built-in" functions that don't provide a signature (such that
# ranged_quantity_input fails)


# This is a wrapper class to expose the ppstruct members as attributes
# (unfortunately, one cannot do dynamical attributes on cdef-classes)
class PathProp(cyprop._PathProp):
    '''
    Container class that holds all path profile properties.

    Parameters
    ----------
    freq : `~astropy.units.Quantity`
        Frequency of radiation [GHz]
    temperature : `~astropy.units.Quantity`
        Ambient temperature at path midpoint [K]
    pressure : `~astropy.units.Quantity`
        Ambient pressure at path midpoint  [hPa]
    lon_t, lat_t : `~astropy.units.Quantity`
        Geographic longitude/latitude of transmitter [deg]
    lon_r, lat_r : `~astropy.units.Quantity`
        Geographic longitude/latitude of receiver [deg]
    h_tg, h_rg : `~astropy.units.Quantity`
        Transmitter/receiver height over ground [m]
    hprof_step : `~astropy.units.Quantity`
        Distance resolution of height profile along path [m]
    timepercent : `~astropy.units.Quantity`
        Time percentage [%] (maximal 50%)
    omega : `~astropy.units.Quantity`, optional
        Fraction of the path over water [%] (see Table 3)
        (default: 0%)
    d_tm : `~astropy.units.Quantity`, optional
        longest continuous land (inland + coastal) section of the
        great-circle path [km]
        (default: distance between Tx and Rx)
    d_lm : `~astropy.units.Quantity`, optional
        longest continuous inland section of the great-circle path [km]
        (default: distance between Tx and Rx)
    d_ct, d_cr : `~astropy.units.Quantity`, optional
        Distance over land from transmitter/receiver antenna to the coast
        along great circle interference path [km]
        (default: 50000 km)
    zone_t, zone_r : CLUTTER enum, optional
        Clutter type for transmitter/receiver terminal.
        (default: CLUTTER.UNKNOWN)
    polarization : int, optional
        Polarization (default: 0)
        Allowed values are: 0 - horizontal, 1 - vertical
    version : int, optional
        ITU-R Rec. P.452 version. Allowed values are: 14, 16
    delta_N : `~astropy.units.Quantity`, optional
        Average radio-refractive index lapse-rate through the lowest 1 km of
        the atmosphere [N-units/km = 1/km]
        (default: query `~pycraf.pathprof.deltaN_N0_from_map`)
    N_0 : `~astropy.units.Quantity`, optional
        Sea-level surface refractivity [N-units = dimless]
        (default: query `~pycraf.pathprof.deltaN_N0_from_map`)
    hprof_dists : `~astropy.units.Quantity`, optional
        Distance vector associated with the height profile `hprof_heights`.
        (default: query `~pycraf.pathprof.srtm_height_profile`)
    hprof_heights : `~astropy.units.Quantity`, optional
        Terrain heights profile for the distances in `hprof_dists`.
        (default: query `~pycraf.pathprof.srtm_height_profile`)
    hprof_bearing : `~astropy.units.Quantity`, optional
        Start bearing of the height profile path.
        (default: query `~pycraf.pathprof.srtm_height_profile`)
    hprof_backbearing : `~astropy.units.Quantity`, optional
        Back-bearing of the height profile path.
        (default: query `~pycraf.pathprof.srtm_height_profile`)

    Returns
    -------
    pprop : PathProp instance

    Notes
    -----
    - The diffraction-loss algorithm was changed between ITU-R P.452
      version 14 and 15. The former used a Deygout method, the new one
      is based on a Bullington calculation with correction terms.

    - Set `d_ct` and `d_cr` to zero for a terminal on ship or on a sea
      platform; only relevant if less than 5 km.

    - Per default, the values for `delta_N` and `N_0` are queried from
      a radiometeorological map provided with `ITU-R Rec. P.452
      <https://www.itu.int/rec/R-REC-P.452-16-201507-I/en>`_
      using the function `~pycraf.pathprof.deltaN_N0_from_map`. If
      you want to use your own values, you have to provide both,
      `delta_N` and `N_0`.

    - Per default, the height-profile data are queried from
      `SRTM data <https://www2.jpl.nasa.gov/srtm/>`_ using the
      `~pycraf.pathprof.srtm_height_profile` function. If
      you want to use your own values, you have to provide all four
      parameters: `hprof_dists`, `hprof_heights`, `bearing`, and
      `back_bearing`.

      If you *don't do* the automatic query from SRTM data, make sure that
      the first element should is zero (transmitter location) and the last
      element is the distance between Tx and Rx. Also, the given
      `lon_t`, `lat_t` and `lon_r`, `lat_r` values should be consistent
      with the height profile. The bearings can be set to zero, if you
      don't need to calculate boresight angles.

      If you *do* the automatic query from
      `SRTM data <https://www2.jpl.nasa.gov/srtm/>`_ you need to downloaded
      them first. Also, an environment variable `SRTMDATA` has to be
      set to point to the directory containing the .hgt files; see
      :ref:`srtm_data`.
    '''

    @utils.ranged_quantity_input(
        freq=(0.1, 100, apu.GHz),
        temperature=(None, None, apu.K),
        pressure=(None, None, apu.hPa),
        lon_t=(-180, 180, apu.deg),
        lat_t=(-90, 90, apu.deg),
        lon_r=(-180, 180, apu.deg),
        lat_r=(-90, 90, apu.deg),
        h_tg=(None, None, apu.m),
        h_rg=(None, None, apu.m),
        hprof_step=(None, None, apu.m),
        timepercent=(0, 50, apu.percent),
        omega=(0, 100, apu.percent),
        d_tm=(None, None, apu.m),
        d_lm=(None, None, apu.m),
        d_ct=(None, None, apu.m),
        d_cr=(None, None, apu.m),
        delta_N=(None, None, cnv.dimless / apu.km),
        N0=(None, None, cnv.dimless),
        hprof_dists=(None, None, apu.km),
        hprof_heights=(None, None, apu.m),
        bearing=(None, None, apu.deg),
        back_bearing=(None, None, apu.deg),
        strip_input_units=True, allow_none=True, output_unit=None
        )
    def __init__(
            self,
            freq,
            temperature,
            pressure,
            lon_t, lat_t,
            lon_r, lat_r,
            h_tg, h_rg,
            hprof_step,
            timepercent,
            omega=0 * apu.percent,
            d_tm=None, d_lm=None,
            d_ct=None, d_cr=None,
            zone_t=cyprop.CLUTTER.UNKNOWN, zone_r=cyprop.CLUTTER.UNKNOWN,
            polarization=0,
            version=16,
            # override if you don't want builtin method:
            delta_N=None, N0=None,
            # override if you don't want builtin method:
            hprof_dists=None, hprof_heights=None,
            hprof_bearing=None, hprof_backbearing=None,
            ):

        super().__init__(
            freq,
            temperature,
            pressure,
            lon_t, lat_t,
            lon_r, lat_r,
            h_tg, h_rg,
            hprof_step,
            timepercent,
            omega=omega,
            d_tm=d_tm, d_lm=d_lm,
            d_ct=d_ct, d_cr=d_cr,
            zone_t=zone_t, zone_r=zone_r,
            polarization=polarization,
            version=version,
            delta_N=delta_N, N0=N0,
            hprof_dists=hprof_dists,
            hprof_heights=hprof_heights,
            hprof_bearing=hprof_bearing,
            hprof_backbearing=hprof_backbearing,
            )

        self.__params = list(cyprop.PARAMETERS_BASIC)  # make a copy
        if self._pp['version'] == 14:
            self.__params += cyprop.PARAMETERS_V14
        elif self._pp['version'] == 16:
            self.__params += cyprop.PARAMETERS_V16

        # no need to set property, as readonly and immutable
        # can just copy to __dict__
        # for p in self.__params:
        #     setattr(
        #         PathProp,
        #         p[0],
        #         property(lambda self: getattr(self._pp, p[0]))
        #         )

        for p in self.__params:
            self.__dict__[p[0]] = self._pp[p[0]] * p[3]

    def __repr__(self):

        return 'PathProp<Freq: {:.3f}>'.format(self.freq)

    def __str__(self):

        return '\n'.join(
            '{}: {{:{}}} {}'.format(
                '{:15s}', p[1], '{:10s}'
                ).format(
                p[0], self._pp[p[0]], p[2]
                )
            for p in self.__params
            )


@utils.ranged_quantity_input(
    output_unit=(cnv.dB, cnv.dB, cnv.dB)
    )
def loss_freespace(pathprop):
    '''
    Calculate the free-space loss, L_bfsg, of a propagating radio wave
    according to ITU-R P.452-16 Eq (8-12).

    Parameters
    ----------
    pathprop : `~pycraf.pathprof.PathProp` instance
        This helper class works as a container to hold various properties
        of the path (e.g., geometry).

    Returns
    -------
    L_bfsg : `~astropy.units.Quantity`
        Free-space loss [dB]
    E_sp : `~astropy.units.Quantity`
        Focussing/multipath correction for p% [dB]
    E_sbeta : `~astropy.units.Quantity`
        Focussing/multipath correction for beta0% [dB]


    Notes
    -----
    - This function is similar to the `~pycraf.conversions.free_space_loss`
      function but additionally accounts for the atmospheric absorption and
      corrects for focusing and multipath effects.
    - With the return values one can also form::

          L_b0p = L_bfsg + E_sp
          L_b0beta = L_bfsg + E_sbeta

      which is necessary for some steps in the ITU-R P.452 algorithms.
    '''

    return cyprop.free_space_loss_bfsg_cython(pathprop)


@utils.ranged_quantity_input(
    G_t=(None, None, cnv.dBi),
    G_r=(None, None, cnv.dBi),
    strip_input_units=True, output_unit=cnv.dB
    )
def loss_troposcatter(
        pathprop, G_t=0. * cnv.dBi, G_r=0. * cnv.dBi,
        ):
    '''
    Calculate the tropospheric scatter loss, L_bs, of a propagating radio wave
    according to ITU-R P.452-16 Eq (45).

    Parameters
    ----------
    pathprop : `~pycraf.pathprof.PathProp` instance
        This helper class works as a container to hold various properties
        of the path (e.g., geometry).
    G_t, G_r  : `~astropy.units.Quantity`
        Antenna gain (transmitter, receiver) in the direction of the
        horizon(!) along the great-circle interference path [dBi]

    Returns
    -------
    L_bs : `~astropy.units.Quantity`
        Tropospheric scatter loss [dB]
    '''

    return cyprop.tropospheric_scatter_loss_bs_cython(pathprop, G_t, G_r)


@utils.ranged_quantity_input(output_unit=cnv.dB)
def loss_ducting(pathprop):
    '''
    Calculate the ducting/layer reflection loss, L_ba, of a propagating radio
    wave according to ITU-R P.452-16 Eq (46-56).

    Parameters
    ----------
    pathprop : `~pycraf.pathprof.PathProp` instance
        This helper class works as a container to hold various properties
        of the path (e.g., geometry).

    Returns
    -------
    L_ba : `~astropy.units.Quantity`
        Ducting/layer reflection loss [dB]
    '''

    return cyprop.ducting_loss_ba_cython(pathprop)


@utils.ranged_quantity_input(
    output_unit=(cnv.dB, cnv.dB, cnv.dB, cnv.dB, cnv.dB)
    )
def loss_diffraction(pathprop):
    '''
    Calculate the Diffraction loss of a propagating radio
    wave according to ITU-R P.452-16 Eq (14-44).

    Parameters
    ----------
    pathprop : `~pycraf.pathprof.PathProp` instance
        This helper class works as a container to hold various properties
        of the path (e.g., geometry).

    Returns
    -------
    L_d_50 : `~astropy.units.Quantity`
        Median diffraction loss [dB]
    L_dp : `~astropy.units.Quantity`
        Diffraction loss not exceeded for p% time, [dB]
    L_bd_50 : `~astropy.units.Quantity`
        Median basic transmission loss associated with
        diffraction [dB]::

            L_bd_50 = L_bfsg + L_d50

    L_bd : `~astropy.units.Quantity`
        Basic transmission loss associated with diffraction not
        exceeded for p% time [dB]::

            L_bd = L_b0p + L_dp

    L_min_b0p : `~astropy.units.Quantity`
        Notional minimum basic transmission loss associated with
        LoS propagation and over-sea sub-path diffraction

    Notes
    -----
    - L_d_50 and L_dp are just intermediary values; the complete
      diffraction loss is L_bd_50 or L_bd, respectively (taking into
      account a free-space loss component for the diffraction path)
    '''

    return cyprop.diffraction_loss_complete_cython(pathprop)


@utils.ranged_quantity_input(
    G_t=(None, None, cnv.dBi),
    G_r=(None, None, cnv.dBi),
    strip_input_units=True,
    output_unit=(cnv.dB, cnv.dB, cnv.dB, cnv.dB, cnv.dB, cnv.dB, cnv.dB)
    )
def loss_complete(
        pathprop, G_t=0. * cnv.dBi, G_r=0. * cnv.dBi,
        ):
    '''
    Calculate the total loss of a propagating radio
    wave according to ITU-R P.452-16 Eq (58-64).

    Parameters
    ----------
    pathprop : `~pycraf.pathprof.PathProp` instance
        This helper class works as a container to hold various properties
        of the path (e.g., geometry).
    G_t, G_r  : `~astropy.units.Quantity`
        Antenna gain (transmitter, receiver) in the direction of the
        horizon(!) along the great-circle interference path [dBi]

    Returns
    -------
    L_bfsg : `~astropy.units.Quantity`
        Free-space loss [dB]
    L_bd : `~astropy.units.Quantity`
        Basic transmission loss associated with diffraction not
        exceeded for p% time [dB]::

            L_bd = L_b0p + L_dp

    L_bs : `~astropy.units.Quantity`
        Tropospheric scatter loss [dB]
    L_ba : `~astropy.units.Quantity`
        Ducting/layer reflection loss [dB]
    L_b : `~astropy.units.Quantity`
        Complete path propagation loss [dB]
    L_b_corr : `~astropy.units.Quantity`
        As L_b but with clutter correction [dB]
    L : `~astropy.units.Quantity`
        As L_b_corr but with gain and clutter correction [dB]
    '''

    return cyprop.path_attenuation_complete_cython(pathprop, G_t, G_r)


@utils.ranged_quantity_input(
    h_g=(None, None, apu.m),
    freq=(None, None, apu.GHz),
    strip_input_units=True,
    output_unit=cnv.dB
    )
def clutter_correction(
        h_g, zone, freq
        ):
    '''
    Calculate the Clutter loss of a propagating radio
    wave according to ITU-R P.452-16 Eq (57).

    Parameters
    ----------
    h_g : `~astropy.units.Quantity`
        Height over ground [m]
    zone : CLUTTER enum, optional
        Clutter category for terminal.
    freq : `~astropy.units.Quantity`
        Frequency of radiation [GHz]

    Returns
    -------
    A_h : `~astropy.units.Quantity`
        Clutter correction to path attenuation [dB]
    '''

    return cyprop.clutter_correction_cython(h_g, zone, freq)


def _clutter_imt(
        freq,
        dist,
        location_percent,
        num_end_points=1,
        ):

    assert num_end_points in [1, 2]

    def Qinv(x):
        # Note, this is *not* identical to cyprop._I_helper
        # only good between 1.e-6 and 0.5
        # See R-Rec P.1546

        x = np.atleast_1d(x).copy()
        mask = x > 0.5
        x[mask] = 1 - x[mask]

        T = np.sqrt(-2 * np.log(x))
        Z = (
            (
                ((0.010328 * T + 0.802853) * T) + 2.515516698
                ) /
            (
                ((0.001308 * T + 0.189269) * T + 1.432788) * T + 1.
                )
            )

        Q = T - Z
        Q[mask] *= -1
        return Q

    # def Qinv(x):
    #     # larger x range than the approximation given in P.1546?
    #     # definitely much slower

    #     from scipy.stats import norm as qnorm

    #     x = np.atleast_1d(x).copy()

    #     mask = x > 0.5
    #     x[mask] = 1 - x[mask]

    #     Q = -qnorm.ppf(x, 0)
    #     Q[mask] *= -1

    #     return Q

    L_l = 23.5 + 9.6 * np.log10(freq)
    L_s = 32.98 + 23.9 * np.log10(dist) + 3.0 * np.log10(freq)

    L_clutter = -5 * np.log10(
        np.power(10, -0.2 * L_l) + np.power(10, -0.2 * L_s)
        ) - 6 * Qinv(location_percent / 100.)

    if num_end_points == 2:
        L_clutter *= 2

    return L_clutter


@utils.ranged_quantity_input(
    freq=(2, 67, apu.GHz),
    dist=(0.25, None, apu.km),
    location_percent=(0, 100, apu.percent),
    strip_input_units=True, output_unit=cnv.dB
    )
def clutter_imt(
        freq,
        dist,
        location_percent,
        num_end_points=1,
        ):
    '''
    Calculate the Clutter loss according to IMT.CLUTTER document (method 2).

    Parameters
    ----------
    freq : `~astropy.units.Quantity`
        Frequency of radiation [GHz]
    dist : `~astropy.units.Quantity`
        Distance between Tx/Rx antennas [km]

        Minimal distance must be 0.25 km (single endpoint clutter) or 1 km
        (if both endpoints are to be corrected for clutter)
    location_percent : `~astropy.units.Quantity`
        Percentage of locations for which the clutter loss `L_clutter`
        (calculated with this function) will not be exceeded [%]
    num_end_points : int, optional
        number of endpoints affected by clutter, allowed values: 1, 2

    Returns
    -------
    L_clutter : `~astropy.units.Quantity`
        Clutter loss [dB]

    Notes
    -----
    - The algorithm is independent of effective antenna height (w.r.t.
      clutter height), i.e., it doesn't distinguish between terminals which
      are close to the ground and those closer to the top of the building.
      However, the model is only appropriate if the terminal is "in the
      clutter", below the rooftops.
    - The result of this function is to be understood as a cumulative
      value. For example, if `location_percent = 2%`, it means that for
      2% of all possible locations, the clutter loss will not exceed the
      returned `L_clutter` value, for the remaining 98% of locations it
      will therefore be lower than `L_clutter`. The smaller `location_percent`,
      the smaller the returned `L_clutter`, i.e., low clutter attenuations
      are more unlikely.
    - This model was proposed by ITU study group SG 3 to replace
      `~pycraf.pathprof.clutter_correction` for IMT 5G studies (especially,
      at higher frequencies, where multipath effects play a role in
      urban and suburban areas).
    '''

    return _clutter_imt(
        freq,
        dist,
        location_percent,
        num_end_points=num_end_points,
        )


# TODO: do we want to convert output dictionary arrays to quantities?
@utils.ranged_quantity_input(
    lon_t=(-180, 180, apu.deg),
    lat_t=(-90, 90, apu.deg),
    map_size_lon=(0.002, 90, apu.deg),
    map_size_lat=(0.002, 90, apu.deg),
    map_resolution=(0.0001, 0.1, apu.deg),
    d_tm=(None, None, apu.m),
    d_lm=(None, None, apu.m),
    d_ct=(None, None, apu.m),
    d_cr=(None, None, apu.m),
    strip_input_units=True, allow_none=True, output_unit=None
    )
def height_profile_data(
        lon_t, lat_t,
        map_size_lon, map_size_lat,
        map_resolution=3. * apu.arcsec,
        do_cos_delta=True,
        zone_t=cyprop.CLUTTER.UNKNOWN, zone_r=cyprop.CLUTTER.UNKNOWN,
        d_tm=None, d_lm=None,
        d_ct=None, d_cr=None,
        ):

    '''
    Calculate height profiles and auxillary maps needed for
    `~pycraf.pathprof.atten_map_fast`.

    This can be used to cache height-profile data. Since it is independent
    of frequency, timepercent, Tx and Rx heights, etc., one can re-use
    it to save computing time when doing batch jobs.

    The basic idea is to only calculate the height profiles between each of
    the pixels on the map edges (super-sampled with a factor of ~3) and the
    map center. The `me` height profiles have different lengths, because
    the distance to the pixels on the edges varies (e.g., corners have the
    largest distance from the center). For shorter profiles, the arrays are
    zero-padded such that all `me` profiles have the same length `mh`. They
    are stored in a 2D array, called `height_profs`. Associated with this,
    a 1D array, `dist_prof`, contains the associated path distances. It has
    length `mh`. For convenience, also a zero-valued array, `zheight_prof`
    is stored, having length `mh`. This is used for the flat-Earth
    calculations in P.452.

    Now, for each other pixel in the map, we have to find out, which of the
    above height profile paths comes closest (with one of its elements) to
    the pixel in question. The identified path ID is stored in `path_idx_map`,
    and the index of the distance element in that path is kept in
    `dist_end_idx_map`. For debugging reasons, the angular distance of
    the "best" path is provided in `pix_dist_map`, but this is not used
    in `~pycraf.pathprof.atten_map_fast`.

    Parameters
    ----------
    lon_t, lat_t : `~astropy.units.Quantity`
        Geographic longitude/latitude of transmitter [deg]
    map_size_lon, map_size_lat : `~astropy.units.Quantity`
        Map size in longitude/latitude[deg]
    map_resolution : `~astropy.units.Quantity`, optional
        Pixel resolution of map [deg] (default: 3 arcsec)
    do_cos_delta : bool, optional
        If True, divide `map_size_lon` by `cos(lat_t)` to produce a more
        square-like map. (default: True)
    zone_t, zone_r : CLUTTER enum, optional
        Clutter type for transmitter/receiver terminal.
        (default: CLUTTER.UNKNOWN)
    d_tm : `~astropy.units.Quantity`, optional
        longest continuous land (inland + coastal) section of the
        great-circle path [km]
        (default: distance between Tx and Rx)
    d_lm : `~astropy.units.Quantity`, optional
        longest continuous inland section of the great-circle path [km]
        (default: distance between Tx and Rx)
    d_ct, d_cr : `~astropy.units.Quantity`, optional
        Distance over land from transmitter/receiver antenna to the coast
        along great circle interference path [km]
        (default: 50000 km)

    Returns
    -------
    hprof_data : dict
        Dictionary with height profiles and auxillary data as
        calculated with `~pycraf.pathprof.height_profile_data`.

        The dictionary contains the following entities (the map dimension
        is mx * my):

        - "lon_t", "lat_t" : float

          Map center coordinates.

        - "map_size_lon", "map_size_lat" : float

          Map size.

        - "hprof_step" : float

          Distance resolution of height profile.

        - "do_cos_delta" : int

          Whether cos-delta correction was applied for map size creation.

        - "xcoords", "ycoords" : `~numpy.ndarray` 1D (float; (mx, ); (my, ))

          Longitude and latitude coordinates of first row and first column
          in the map

        - "lon_mid_map", "lat_mid_map" : `~numpy.ndarray` 2D (float; (mx, my))

          Longitude and latitude path center coordinates for each pixel
          w.r.t. map center.

        - "dist_map" : `~numpy.ndarray` 2D (float; (mx, my))

          Distances to map center for each pixel.

        - "d_ct_map", "d_cr_map" : `~numpy.ndarray` 2D (float; (mx, my))

          The `d_ct` and `d_cr` values for each pixel in the map.

        - "d_lm_map", "d_tm_map" : `~numpy.ndarray` 2D (float; (mx, my))

          The `d_lm` and `d_tm` values for each pixel in the map.

        - "zone_t_map", "zone_r_map" : `~numpy.ndarray` 2D (CLUTTER enum; (mx, my))

          The clutter zone types `zone_t` and `zone_r` for each pixel in the
          map.

        - "bearing_map", "back_bearing_map" : `~numpy.ndarray` 2D (float; (mx, my))

          The `bearing` and `backbearing` values for each pixel in the map.

        - "N0_map", "delta_N_map", "beta0_map" : `~numpy.ndarray` 2D (float; (mx, my))

          The `N0`, `delta_N`, and `beta0` values for each pixel in the map.

        - "path_idx_map" : `~numpy.ndarray` 2D (int; (mx, my))

          Path IDs for each pixel in the map. With this index, one can query
          the associated height profile from `height_profs`.

        - "pix_dist_map" : `~numpy.ndarray` 2D (float; (mx, my))

          Angular distance of the closest path to each of the map pixels.

        - "dist_end_idx_map" : `~numpy.ndarray` 2D (int; (mx, my))

          Index of the last element in the dist/height profiles to be used
          when querying the profiles from `dist_prof` and
          `height_profs`.

        - "dist_prof" : `~numpy.ndarray` 1D (float, (mh, ))

          Distance values for each of the paths stored in `height_profs`.

        - "height_profs" : `~numpy.ndarray` 2D (float, (me, mh))

          Height profiles to each of the pixels on the map edge, zero padded.

        - "zheight_prof" : `~numpy.ndarray` 1D (float, (mh, ))

          Zero-valued array of the same length as `height_profs` for
          convenience.

    Notes
    -----
    - Path attenuation is completely symmetric, i.e., it doesn't matter if
      the transmitter or the receiver is situated in the map center.
    - `SRTM data <https://www2.jpl.nasa.gov/srtm/>`_ need to be downloaded
      manually by the user. An environment variable `SRTMDATA` has to be
      set to point to the directory containing the .hgt files; see
      :ref:`srtm_data`.
    '''

    return cyprop.height_profile_data_cython(
        lon_t, lat_t,
        map_size_lon, map_size_lat,
        map_resolution=map_resolution,
        do_cos_delta=1 if do_cos_delta else 0,
        zone_t=zone_t, zone_r=zone_r,
        d_tm=d_tm, d_lm=d_lm,
        d_ct=d_ct, d_cr=d_cr,
        )


@utils.ranged_quantity_input(
    freq=(0.1, 100, apu.GHz),
    temperature=(None, None, apu.K),
    pressure=(None, None, apu.hPa),
    h_tg=(None, None, apu.m),
    h_rg=(None, None, apu.m),
    timepercent=(0, 50, apu.percent),
    omega_percent=(0, 100, apu.percent),
    strip_input_units=True,
    output_unit=(cnv.dB, apu.deg, apu.deg)
    )
def atten_map_fast(
        freq,
        temperature,
        pressure,
        h_tg, h_rg,
        timepercent,
        hprof_data,  # dict_like
        omega=0 * apu.percent,
        polarization=0,
        version=16,
        ):
    '''
    Calculate attenuation maps using a fast method.

    Parameters
    ----------
    freq : `~astropy.units.Quantity`
        Frequency of radiation [GHz]
    temperature : `~astropy.units.Quantity`
        Temperature (K)
    pressure : `~astropy.units.Quantity`
        Pressure (hPa)
    h_tg, h_rg : `~astropy.units.Quantity`
        Transmitter/receiver heights over ground [m]
    timepercent : `~astropy.units.Quantity`
        Time percentage [%] (maximal 50%)
    hprof_data : dict, dict-like
        Dictionary with height profiles and auxillary maps as
        calculated with `~pycraf.pathprof.height_profile_data`.
    omega : `~astropy.units.Quantity`, optional
        Fraction of the path over water [%] (see Table 3)
        (default: 0%)
    polarization : int, optional
        Polarization (default: 0)
        Allowed values are: 0 - horizontal, 1 - vertical
    version : int, optional
        ITU-R Rec. P.452 version. Allowed values are: 14, 16

    Returns
    -------
    atten_map : 3D `~numpy.ndarray`
        Attenuation maps. First dimension has length 6, which refers to:

        0) L_bfsg - Free-space loss [dB]
        1) L_bd - Basic transmission loss associated with diffraction
           not exceeded for p% time [dB]; L_bd = L_b0p + L_dp
        2) L_bs - Tropospheric scatter loss [dB]
        3) L_ba - Ducting/layer reflection loss [dB]
        4) L_b - Complete path propagation loss [dB]
        5) L_b_corr - As L_b but with clutter correction [dB]

        (i.e., the output of path_attenuation_complete without
        gain-corrected values)
    eps_pt_map : 2D `~numpy.ndarray`
        Elevation angle of paths w.r.t. Tx [deg]
    eps_pr_map : 2D `~numpy.ndarray`
        Elevation angle of paths w.r.t. Rx [deg]

    Notes
    -----
    - The diffraction-loss algorithm was changed between ITU-R P.452
      version 14 and 15. The former used a Deygout method, the new one
      is based on a Bullington calculation with correction terms.
    '''

    return cyprop.atten_map_fast_cython(
        freq,
        temperature,
        pressure,
        h_tg, h_rg,
        timepercent,
        hprof_data,  # dict_like
        omega=omega,
        polarization=polarization,
        version=version,
        )


if __name__ == '__main__':
    print('This not a standalone python program! Use as module.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import (
    absolute_import, unicode_literals, division, print_function
    )

# from functools import partial, lru_cache
import os
from astropy import units as apu
import numpy as np
import matplotlib
from scipy.interpolate import RegularGridInterpolator
from .. import conversions as cnv
from .. import utils


__all__ = [
    'R_E', 'K_BETA', 'A_BETA',
    'annual_timepercent_from_worst_month',
    'deltaN_N0_from_map', 'radiomet_data_for_pathcenter',
    'eff_earth_radius_factor_median',
    'eff_earth_radius_factor_beta',
    'eff_earth_radius_median', 'eff_earth_radius_beta',
    'make_kmz', 'terrain_cmap_factory',
    ]


# useful constants
R_E_VALUE = 6371.
R_E = R_E_VALUE * apu.km  # Earth radius
R_E.__doc__ = '''Earth Radius'''

K_BETA_VALUE = 3.
K_BETA = K_BETA_VALUE * cnv.dimless  # eff. Earth radius factor for beta_0
K_BETA.__doc__ = '''Effective Earth radius factor for beta_0 percent'''

A_BETA_VALUE = 3. * R_E_VALUE
A_BETA = K_BETA_VALUE * apu.km  # eff. Earth radius for beta_0
A_BETA.__doc__ = '''Effective Earth radius for beta_0 percent'''

KML_TEMPLATE = '''<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
    <Folder>
    <name>Attenuation map</name>
    <description>Results from pycraf package</description>
    <GroundOverlay>
        <name>Attenuation map on terrain</name>
        <description>Results from the pycraf package</description>
        <color>aaffffff</color>
        <Icon><href>pycraf_atten_map_kmz.png</href></Icon>
        <LatLonBox>
            <east>{:.6f}</east>
            <south>{:.6f}</south>
            <west>{:.6f}</west>
            <north>{:.6f}</north>
        </LatLonBox>
    </GroundOverlay>
    </Folder>
</kml>
'''

# maybe, the following should only be computed on demand?
this_dir, this_filename = os.path.split(__file__)
fname_refract_data = os.path.join(
    this_dir, '../itudata/p.452-16', 'refract_map.npz'
    )
_refract_data = np.load(fname_refract_data)


_DN_interpolator = RegularGridInterpolator(
    (_refract_data['lons'][0], _refract_data['lats'][::-1, 0]),
    _refract_data['dn50'][::-1].T
    )
_N0_interpolator = RegularGridInterpolator(
    (_refract_data['lons'][0], _refract_data['lats'][::-1, 0]),
    _refract_data['n050'][::-1].T
    )


@utils.ranged_quantity_input(
    p_w=(0, 100, apu.percent),
    phi=(-90, 90, apu.deg),
    omega=(0, 100, apu.percent),
    strip_input_units=True, output_unit=apu.percent,
    )
def annual_timepercent_from_worst_month(
        p_w, phi, omega
        ):
    '''
    Calculate annual equivalent time percentage, p, from worst-month time
    percentage, p_w, according to ITU-R P.452-16 Eq (1).

    Parameters
    ----------
    p_w : `~astropy.units.Quantity`
        worst-month time percentage [%]
    phi : `~astropy.units.Quantity`
        Geographic latitude of path center [deg]
    omega : `~astropy.units.Quantity`
        Fraction of the path over water (see Table 3) [%]

    Returns
    -------
    p : `~astropy.units.Quantity`
        Annual equivalent time percentage [%]

    Notes
    -----
    - Use this function, if you want to do path propagation calculations
      for the worst month case. The resulting time percentage, p, can then
      be plugged into other functions. If you want just annual averages,
      simply use your time percentage value as is.
    '''

    omega /= 100.  # convert from percent to fraction

    tmp = np.abs(np.cos(2 * np.radians(phi))) ** 0.7
    G_l = np.sqrt(
        np.where(np.abs(phi) <= 45, 1.1 + tmp, 1.1 - tmp)
        )

    a = np.log10(p_w) + np.log10(G_l) - 0.186 * omega - 0.444
    b = 0.078 * omega + 0.816
    p = 10 ** (a / b)

    p = np.max([p, p_w / 12.], axis=0)

    return p


def _DN_N0_from_map(lon, lat):

    _DN = _DN_interpolator((lon % 360, lat))
    _N0 = _N0_interpolator((lon % 360, lat))

    return _DN, _N0


@utils.ranged_quantity_input(
    lon=(0, 360, apu.deg),
    lat=(-90, 90, apu.deg),
    strip_input_units=True,
    output_unit=(cnv.dimless / apu.km, apu.percent, cnv.dimless),
    )
def deltaN_N0_from_map(lon, lat):
    '''
    Query delta_N and N_0 values from digitized maps by means of bilinear
    interpolation.

    Parameters
    ----------
    lon, lat : `~astropy.units.Quantity`
        Geographic longitude and latitude of path center [deg]

    Returns
    -------
    delta_N : `~astropy.units.Quantity`
        Average radio-refractive index lapse-rate through the lowest 1 km of
        the atmosphere [N-units/km == 1/km]
    N_0 : `~astropy.units.Quantity`
        Sea-level surface refractivity [N-units == dimless]

    Notes
    -----
    - The values for `delta_N` and `N_0` are queried from
      a radiometeorological map provided with `ITU-R Rec. P.452
      <https://www.itu.int/rec/R-REC-P.452-16-201507-I/en>`_.
    '''

    return _DN_N0_from_map(lon, lat)


def _radiomet_data_for_pathcenter(lon, lat, d_tm, d_lm):
    _DN = _DN_interpolator((lon % 360, lat))
    _N0 = _N0_interpolator((lon % 360, lat))

    _tau = 1. - np.exp(-4.12e-4 * np.power(d_lm, 2.41))
    _absphi = np.abs(lat)

    _a = np.power(10, -d_tm / (16. - 6.6 * _tau))
    _b = np.power(10, -5 * (0.496 + 0.354 * _tau))
    _mu1 = np.power(_a + _b, 0.2)
    _mu1 = np.where(_mu1 <= 1, _mu1, 1.)
    _log_mu1 = np.log10(_mu1)

    _phi_cond = _absphi <= 70.
    _mu4 = np.where(
        _phi_cond,
        np.power(10, (-0.935 + 0.0176 * _absphi) * _log_mu1),
        np.power(10, 0.3 * _log_mu1)
        )

    beta_0 = np.where(
        _phi_cond,
        np.power(10, -0.015 * _absphi + 1.67) * _mu1 * _mu4,
        4.17 * _mu1 * _mu4
        )

    return _DN, beta_0, _N0


@utils.ranged_quantity_input(
    lon=(0, 360, apu.deg),
    lat=(-90, 90, apu.deg),
    d_tm=(0, None, apu.km),
    d_lm=(0, None, apu.km),
    strip_input_units=True,
    output_unit=(cnv.dimless / apu.km, apu.percent, cnv.dimless),
    )
def radiomet_data_for_pathcenter(lon, lat, d_tm, d_lm):
    '''
    Calculate delta_N, beta_0, and N_0 values from digitized maps, according
    to ITU-R P.452-16 Eq (2-4).

    Parameters
    ----------
    lon, lat : `~astropy.units.Quantity`
        Geographic longitude and latitude of path center [deg]
    d_tm : `~astropy.units.Quantity`, optional
        longest continuous land (inland + coastal) section of the
        great-circle path [km]
    d_lm : `~astropy.units.Quantity`, optional
        longest continuous inland section of the great-circle path [km]

    Returns
    -------
    delta_N : `~astropy.units.Quantity`
        Average radio-refractive index lapse-rate through the lowest 1 km of
        the atmosphere [N-units/km == 1/km]
    beta_0 : `~astropy.units.Quantity`
        the time percentage for which refractive index lapse-rates
        exceeding 100 N-units/km can be expected in the first 100 m
        of the lower atmosphere [%]
    N_0 : `~astropy.units.Quantity`
        Sea-level surface refractivity [N-units == dimless]

    Notes
    -----
    - The values for `delta_N` and `N_0` are queried from
      a radiometeorological map provided with `ITU-R Rec. P.452
      <https://www.itu.int/rec/R-REC-P.452-16-201507-I/en>`_.
    - Radio-climaticzones can be obtained from
      `ITU Digitized World Map (IDWM) <http://www.itu.int/pub/R-SOFT-IDWM>`_.
      For many applications, it is probably the case, that only inland
      zones are present along the path of length d.
      In this case, d_tm = d_lm = d.
    '''

    return _radiomet_data_for_pathcenter(lon, lat, d_tm, d_lm)


@utils.ranged_quantity_input(
    lon=(0, 360, apu.deg),
    lat=(-90, 90, apu.deg),
    strip_input_units=True,
    output_unit=cnv.dimless,
    )
def eff_earth_radius_factor_median(lon, lat):
    '''
    Calculate median effective Earth radius factor, k_50, according to
    ITU-R P.452-16 Eq (5).

    Parameters
    ----------
    lon, lat : `~astropy.units.Quantity`
        Geographic longitude and latitude of path center [deg]

    Returns
    -------
    k50 : `~astropy.units.Quantity`
        Median effective Earth radius factor [dimless]

    Notes
    -----
    - Uses delta_N, which is derived from digitized maps (shipped with P.452)
      by bilinear interpolation; see also
      `~pycraf.pathprof.deltaN_N0_from_map`.
    '''

    return 157. / (157. - _DN_interpolator((lon % 360, lat)))


@utils.ranged_quantity_input(
    strip_input_units=True,
    output_unit=cnv.dimless,
    )
def eff_earth_radius_factor_beta():
    '''
    Calculate effective Earth radius factor exceeded for beta_0 percent
    of time, k_beta, according to ITU-R P.452-16.

    Returns
    -------
    k_beta : `~astropy.units.Quantity`
        Effective Earth radius factor exceeded for beta_0 percent
        of time [dimless]

    Notes
    -----
    - This is just a constant. You could also use K_BETA to avoid overhead.
    '''

    return K_BETA_VALUE


def _eff_earth_radius_median(lon, lat):

    return R_E_VALUE * 157. / (157. - _DN_interpolator((lon % 360, lat)))


@utils.ranged_quantity_input(
    lon=(0, 360, apu.deg),
    lat=(-90, 90, apu.deg),
    strip_input_units=True,
    output_unit=apu.km,
    )
def eff_earth_radius_median(lon, lat):
    '''
    Calculate median effective Earth radius, a_e, according to
    ITU-R P.452-16 Eq (6a).

    Parameters
    ----------
    lon, lat : `~astropy.units.Quantity`
        Geographic longitude and latitude of path center [deg]

    Returns
    -------
    a_e : `~astropy.units.Quantity`
        Median effective Earth radius [km]

    Notes
    -----
    - Uses delta_N, which is derived from digitized maps (shipped with P.452)
      by bilinear interpolation; see also
      `~pycraf.pathprof.deltaN_N0_from_map`.
    '''

    return _eff_earth_radius_median(lon, lat)


@utils.ranged_quantity_input(
    strip_input_units=True,
    output_unit=apu.km,
    )
def eff_earth_radius_beta(lon, lat):
    '''
    Calculate effective Earth radius exceeded for beta_0 percent of time,
    a_beta, according to ITU-R P.452-16 Eq (6b).

    Returns
    -------
    a_beta : `~astropy.units.Quantity`
        Effective Earth radius exceeded for beta_0 percent of time [km]

    Notes
    -----
    - This is just a constant. You could also use A_BETA to avoid overhead.
    '''

    return A_BETA_VALUE


def make_kmz(
        kmz_filename, atten_map, bbox, vmin=None, vmax=None, cmap='inferno_r'
        ):
    '''
    Produce kmz file for use in GIS software (e.g., Google Earth).

    Parameters
    ----------
    kmz_filename : str
        Output file name for .kmz-file
    atten_map : 2D `~numpy.ndarray` of floats
        2D array with path attenuation values
    bbox : tuple of 4 floats
        (east, south, west, north) edges of map [deg]
    vmin, vmax : float
        Lower and upper colorbar bounds.
        If None, 2.5% and 97.5% percentiles of atten_map are used
        (default: None)
    cmap : matplotlib.colormap
        (default: 'inferno_r')
    '''

    # descriptive xml
    kml = KML_TEMPLATE.format(*bbox)

    # produce jpg, use python pillow for this;
    # however, we don't want this as global requirement,
    # therefore, just a local import
    # from PIL import Image
    # from matplotlib import colors, cm

    if vmin is None:
        vmin = np.percentile(atten_map.flatten(), 2.5)

    if vmax is None:
        vmax = np.percentile(atten_map.flatten(), 97.5)

    # norm = colors.Normalize(vmin=vmin, vmax=vmax)
    # csm = cm.ScalarMappable(norm=norm, cmap=cmap)

    # rgba = csm.to_rgba(atten_map)
    # jpeg = Image.fromarray(np.int32(255 * rgba + 0.5), mode='RGBA')

    from matplotlib.image import imsave
    from io import BytesIO

    png_buf = BytesIO()

    imsave(
        png_buf,
        atten_map,
        vmin=vmin, vmax=vmax, cmap=cmap,
        origin='lower'
        )

    png_buf.seek(0)

    # write as kmz (zip file)
    import zipfile

    with zipfile.ZipFile(kmz_filename, 'w') as myzip:
        myzip.writestr('pycraf_atten_map_kmz.png', png_buf.read())
        myzip.writestr('doc.kml', kml)


class FixPointNormalize(matplotlib.colors.Normalize):
    '''
    From http://stackoverflow.com/questions/40895021/python-equivalent-for-matlabs-demcmap-elevation-appropriate-colormap
    by ImportanceOfBeingErnest

    Inspired by http://stackoverflow.com/questions/20144529/shifted-colorbar-matplotlib
    Subclassing Normalize to obtain a colormap with a fixpoint
    somewhere in the middle of the colormap.

    This may be useful for a `terrain` map, to set the 'sea level'
    to a color in the blue/turquise range.
    '''

    def __init__(
            self,
            vmin=None, vmax=None, sealevel=0,
            col_val=0.21875, clip=False
            ):

        # sealevel is the fix point of the colormap (in data units)
        self.sealevel = sealevel
        # col_val is the color value in the range [0, 1]
        # that should represent the sealevel.
        self.col_val = col_val
        matplotlib.colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):

        x, y = [self.vmin, self.sealevel, self.vmax], [0, self.col_val, 1]
        return np.ma.masked_array(np.interp(value, x, y))


def terrain_cmap_factory(sealevel=0.5, vmax=1200):
    '''
    Produce terrain colormap and norm to be used in plt.imshow.

    With this, one can adjust the colors in the cmap such that the sea level
    is properly defined (blue).

    A simple use case would look like the following::

        >>> vmin, vmax = -20, 1200  # doctest: +SKIP
        >>> terrain_cmap, terrain_norm = terrain_cmap_factory(vmax=vmax)  # doctest: +SKIP
        >>> plt.imshow(  # doctest: +SKIP
        ...     heights, cmap=terrain_cmap, norm=terrain_norm,
        ...     vmin=vmin, vmax=vmax
        ...     )

    Parameters
    ----------
    sealevel : float
        The sealevel value.
    vmax : float
        Maximum height to cover in the colormap
        (one should call plt.imshow with the same vmax option!)

    Returns
    -------
    terrain_cmap : matplotlib.colors.LinearSegmentedColormap
    terrain_norm : matplotlib.colors.Normalize instance
    '''

    # Combine the lower and upper range of the terrain colormap with a gap in
    # the middle to let the coastline appear more prominently. Inspired by
    # stackoverflow.com/questions/31051488/combining-two-matplotlib-colormaps

    colors_undersea = matplotlib.pyplot.cm.terrain(np.linspace(0, 0.17, 56))
    colors_land = matplotlib.pyplot.cm.terrain(np.linspace(0.25, 1, 200))

    # combine them and build a new colormap
    colors = np.vstack((colors_undersea, colors_land))
    terrain_cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
        'terrain_normed', colors
        )
    terrain_norm = FixPointNormalize(sealevel=sealevel, vmax=vmax)

    return terrain_cmap, terrain_norm


if __name__ == '__main__':
    print('This not a standalone python program! Use as module.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
Top-level functionality:
'''

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *
# ----------------------------------------------------------------------------

if not _ASTROPY_SETUP_:
    # For egg_info test builds to pass, put package imports here.

    # from .example_mod import *

    from . import antenna
    from . import atm
    from . import conversions
    from . import geospatial
    from . import utils
    from . import pathprof
    from . import protection



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import (
    absolute_import, unicode_literals, division, print_function
    )

from functools import partial, lru_cache
# import numpy as np
import numbers
from astropy import units as apu
import pyproj
from .. import utils


# Note: UTM zones extend over 6d in longitude and a full hemisphere in
# latitude. However, the NATO system further refines the zones by
# putting them into latitude zones, to form a grid. pyproj uses the former
# approach, though. See "https://en.wikipedia.org/wiki/
# Universal_Transverse_Mercator_coordinate_system" for further details.
# In the following, we use "N" and "S" to make a distinction between
# Northern and Southern hemisphere. Don't mix this up with the grid
# nomenclature (where xxS is a cell on the Northern hemisphere...).


__all__ = [
    'utm_to_wgs84', 'wgs84_to_utm',  # 'utm_to_wgs84_32N',
    'etrs89_to_wgs84', 'wgs84_to_etrs89',
    # 'itrf2005_to_wgs84',
    # 'wgs84_to_itrf2005',
    ]


@lru_cache(maxsize=16, typed=True)
def _create_proj(sys1, sys2, zone=None, south=False):
    '''
    Helper function to create and cache pyproj.Proj instances.
    '''

    if sys1 == 'UTM' and sys2 == 'WGS84':

        if not isinstance(zone, numbers.Integral):
            raise TypeError('zone must be an integer')

        _proj_str = (
            '+proj=utm +ellps=WGS84 +datum=WGS84 +units=m +no_defs '
            '+zone={:d}'.format(zone)
            )
        if south:
            _proj_str += '+south'

    elif sys1 == 'ETRS89' and sys2 == 'WGS84':

        _proj_str = (
            '+proj=laea +lat_0=52 +lon_0=10 +x_0=4321000 +y_0=3210000 '
            '+ellps=GRS80 +units=m +no_defs'
            )
    elif sys1 == 'ITRF05' and sys2 == 'WGS84':

        _proj_str = (
            '+proj=longlat +datum=WGS84 +no_defs'
            '+to +proj=geocent +ellps=GRS80 +units=m +no_defs'
            )
    else:
        raise AssertionError(
            'requested conversion not supported {} {}'.format(sys1, sys2)
            )

    return pyproj.Proj(_proj_str)


@utils.ranged_quantity_input(
    ulon=(None, None, apu.m),
    ulat=(None, None, apu.m),
    strip_input_units=True, output_unit=(apu.deg, apu.deg)
    )
def utm_to_wgs84(ulon, ulat, zone, south=False):
    '''
    Convert UTM coordinates to GPS/WGS84.

    Parameters
    ----------
    ulon, ulat : `~astropy.units.Quantity`
        UTM longitude and latitude [m]
    zone : int
        UTM zone (e.g., 32 for longitudes 6 E to 12 E); see `Wikipedia
        <https://en.wikipedia.org/wiki/Universal_Transverse_Mercator_coordinate_system>`_
        for more information on the zones
    south : bool
        Hemisphere; set to True for southern hemisphere (default: False)

    Returns
    -------
    glon, glat : `~astropy.units.Quantity`
        GPS/WGS84 longitude and latitude [deg]

    Notes
    -----
    - Uses

      .. code-block:: bash

          +proj=utm +zone=xx +ellps=WGS84 +datum=WGS84 +units=m +no_defs [+south]

      for `pyproj` setup (inverse transformation). Only one zone per function
      call is allowed.
    - This function uses only the longitudal zone scheme. There is also
      the NATO system, which introduces latitude bands.
    '''

    _proj = _create_proj('UTM', 'WGS84', zone, south)

    return _proj(ulon, ulat, inverse=True)


@utils.ranged_quantity_input(
    glon=(None, None, apu.deg),
    glat=(None, None, apu.deg),
    strip_input_units=True, output_unit=(apu.m, apu.m)
    )
def wgs84_to_utm(glon, glat, zone, south=False):
    '''
    Convert GPS/WGS84 coordinates to UTM.

    Parameters
    ----------
    glon, glat : `~astropy.units.Quantity`
        GPS/WGS84 longitude and latitude [deg]
    zone : int
        UTM zone (e.g., 32 for longitudes 6 E to 12 E); see `Wikipedia
        <https://en.wikipedia.org/wiki/Universal_Transverse_Mercator_coordinate_system>`_
        for more information on the zones
    south : bool
        Hemisphere; set to True for southern hemisphere (default: False)

    Returns
    -------
    ulon, ulat : `~astropy.units.Quantity`
        UTM longitude and latitude [m]

    Notes
    -----
    - Uses

      .. code-block:: bash

          +proj=utm +zone=xx +ellps=WGS84 +datum=WGS84 +units=m +no_defs [+south]

      for `pyproj` setup (forward transformation). Only one zone per function
      call is allowed.
    - This function uses only the longitudal zone scheme. There is also
      the NATO system, which introduces latitude bands.
    '''

    _proj = _create_proj('UTM', 'WGS84', zone, south)

    return _proj(glon, glat, inverse=False)


# This is for Western Germany (Effelsberg)
utm_to_wgs84_32N = partial(utm_to_wgs84, zone=32, south=False)
wgs84_to_utm_32N = partial(wgs84_to_utm, zone=32, south=False)


@utils.ranged_quantity_input(
    elon=(None, None, apu.m),
    elat=(None, None, apu.m),
    strip_input_units=True, output_unit=(apu.deg, apu.deg)
    )
def etrs89_to_wgs84(elon, elat):
    '''
    Convert ETSR89 coordinates to GPS/WGS84.

    ETRS89 is the European Terrestrial Reference System.
    (Using a Lambert Equal Area projection.)

    Parameters
    ----------
    elon, elat : `~astropy.units.Quantity`
        ETRS89 longitude and latitude [m]

    Returns
    -------
    glon, glat : `~astropy.units.Quantity`
        GPS/WGS84 longitude and latitude [deg]

    Notes
    -----
    Uses

    .. code-block:: bash

        +proj=laea +lat_0=52 +lon_0=10 +x_0=4321000 +y_0=3210000
        +ellps=GRS80 +units=m +no_defs

    for pyproj setup (inverse transformation).
    '''

    _proj = _create_proj('ETRS89', 'WGS84')

    return _proj(elon, elat, inverse=True)


@utils.ranged_quantity_input(
    glon=(None, None, apu.deg),
    glat=(None, None, apu.deg),
    strip_input_units=True, output_unit=(apu.m, apu.m)
    )
def wgs84_to_etrs89(glon, glat):
    '''
    Convert GPS/WGS84 coordinates to ETSR89.

    ETRS89 is the European Terrestrial Reference System.
    (Using a Lambert Equal Area projection.)

    Parameters
    ----------
    glon, glat : `~astropy.units.Quantity`
        GPS/WGS84 longitude and latitude [deg]

    Returns
    -------
    elon, elat : `~astropy.units.Quantity`
        ETRS89 longitude and latitude [m]

    Notes
    -----
    Uses

    .. code-block:: bash

        +proj=laea +lat_0=52 +lon_0=10 +x_0=4321000 +y_0=3210000
        +ellps=GRS80 +units=m +no_defs

    for pyproj setup (forward transformation).
    '''

    _proj = _create_proj('ETRS89', 'WGS84')

    return _proj(glon, glat, inverse=False)


def _wgs84_to_itrf2005(glon, glat):
    '''
    BROKEN!

    Convert GPS/WGS84 coordinates to ITRF.

    Parameters
    ----------
    glon, glat - GPS/WGS84 longitude and latitude [deg]

    Returns
    -------
    x, y, z - ITRF cartesian coordinates, geocentric [m]

    Notes
    -----
    Uses
        +proj=utm +zone=xx +ellps=WGS84 +datum=WGS84 +units=m +no_defs [+south]
    for pyproj setup.
    '''

    _proj = _create_proj('ITRF05', 'WGS84')

    return _proj(glon, glat, inverse=False)


if __name__ == '__main__':
    print('This not a standalone python program! Use as module.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
This sub-package contains various protection criteria and interference limits,
e.g. for the radio-astronomy service (RAS).
'''

from .ra769 import *
from .cispr import *



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This directive takes a single argument that must be a module or package. It
will produce a block of documentation that includes the docstring for the
package, an :ref:`automodsumm` directive, and an :ref:`automod-diagram` if
there are any classes in the module. If only the main docstring of the
module/package is desired in the documentation, use `automodule`_ instead of
`automodapi`_.

It accepts the following options:

    * ``:include-all-objects:``
        If present, include not just functions and classes, but all objects.
        This includes variables, for which a possible docstring after the
        variable definition will be shown.

    * ``:no-inheritance-diagram:``
        If present, the inheritance diagram will not be shown even if
        the module/package has classes.

    * ``:skip: str``
        This option results in the
        specified object being skipped, that is the object will *not* be
        included in the generated documentation. This option may appear
        any number of times to skip multiple objects.

    * ``:no-main-docstr:``
        If present, the docstring for the module/package will not be generated.
        The function and class tables will still be used, however.

    * ``:headings: str``
        Specifies the characters (in one string) used as the heading
        levels used for the generated section. This must have at least 2
        characters (any after 2 will be ignored). This also *must* match
        the rest of the documentation on this page for sphinx to be
        happy. Defaults to "-^", which matches the convention used for
        Python's documentation, assuming the automodapi call is inside a
        top-level section (which usually uses '=').

    * ``:no-heading:``
        If specified do not create a top level heading for the section.
        That is, do not create a title heading with text like "packagename
        Package".  The actual docstring for the package/module will still be
        shown, though, unless ``:no-main-docstr:`` is given.

    * ``:allowed-package-names: str``
        Specifies the packages that functions/classes documented here are
        allowed to be from, as comma-separated list of package names. If not
        given, only objects that are actually in a subpackage of the package
        currently being documented are included.

    * ``:inherited-members:`` / ``:no-inherited-members:``
        The global sphinx configuration option
        ``automodsumm_inherited_members`` decides if members that a class
        inherits from a base class are included in the generated
        documentation. The option ``:inherited-members:`` or ``:no-inherited-members:``
        allows the user to overrride the global setting.


This extension also adds three sphinx configuration options:

* ``automodapi_toctreedirnm``
    This must be a string that specifies the name of the directory the
    automodsumm generated documentation ends up in. This directory path should
    be relative to the documentation root (e.g., same place as ``index.rst``).
    Defaults to ``'api'``.

* ``automodapi_writereprocessed``
    Should be a bool, and if `True`, will cause `automodapi`_ to write files
    with any `automodapi`_ sections replaced with the content Sphinx
    processes after `automodapi`_ has run.  The output files are not
    actually used by sphinx, so this option is only for figuring out the
    cause of sphinx warnings or other debugging.  Defaults to `False`.

* ``automodsumm_inherited_members``
    Should be a bool and if ``True`` members that a class inherits from a base
    class are included in the generated documentation. Defaults to ``False``.

.. _automodule: http://sphinx-doc.org/latest/ext/autodoc.html?highlight=automodule#directive-automodule
"""

# Implementation note:
# The 'automodapi' directive is not actually implemented as a docutils
# directive. Instead, this extension searches for the 'automodapi' text in
# all sphinx documents, and replaces it where necessary from a template built
# into this extension. This is necessary because automodsumm (and autosummary)
# use the "builder-inited" event, which comes before the directives are
# actually built.

import inspect
import io
import os
import re
import sys

from .utils import find_mod_objs

if sys.version_info[0] == 3:
    text_type = str
else:
    text_type = unicode


automod_templ_modheader = """
{modname} {pkgormod}
{modhds}{pkgormodhds}

{automoduleline}
"""

automod_templ_classes = """
Classes
{clshds}

.. automodsumm:: {modname}
    :classes-only:
    {clsfuncoptions}
"""

automod_templ_funcs = """
Functions
{funchds}

.. automodsumm:: {modname}
    :functions-only:
    {clsfuncoptions}
"""

automod_templ_vars = """
Variables
{otherhds}

.. automodsumm:: {modname}
    :variables-only:
    {clsfuncoptions}
"""

automod_templ_inh = """
Class Inheritance Diagram
{clsinhsechds}

.. automod-diagram:: {modname}
    :private-bases:
    :parts: 1
    {allowedpkgnms}
    {skip}
"""

_automodapirex = re.compile(r'^(?:\.\.\s+automodapi::\s*)([A-Za-z0-9_.]+)'
                            r'\s*$((?:\n\s+:[a-zA-Z_\-]+:.*$)*)',
                            flags=re.MULTILINE)
# the last group of the above regex is intended to go into finall with the below
_automodapiargsrex = re.compile(r':([a-zA-Z_\-]+):(.*)$', flags=re.MULTILINE)


def automodapi_replace(sourcestr, app, dotoctree=True, docname=None,
                       warnings=True):
    """
    Replaces `sourcestr`'s entries of ".. automdapi::" with the
    automodapi template form based on provided options.

    This is used with the sphinx event 'source-read' to replace
    `automodapi`_ entries before sphinx actually processes them, as
    automodsumm needs the code to be present to generate stub
    documentation.

    Parameters
    ----------
    sourcestr : str
        The string with sphinx source to be checked for automodapi
        replacement.
    app : `sphinx.application.Application`
        The sphinx application.
    dotoctree : bool
        If `True`, a ":toctree:" option will be added in the "..
        automodsumm::" sections of the template, pointing to the
        appropriate "generated" directory based on the Astropy convention
        (e.g. in ``docs/api``)
    docname : str
        The name of the file for this `sourcestr` (if known - if not, it
        can be `None`). If not provided and `dotoctree` is `True`, the
        generated files may end up in the wrong place.
    warnings : bool
        If `False`, all warnings that would normally be issued are
        silenced.

    Returns
    -------
    newstr :str
        The string with automodapi entries replaced with the correct
        sphinx markup.
    """

    spl = _automodapirex.split(sourcestr)
    if len(spl) > 1:  # automodsumm is in this document

        # Use app.srcdir because api folder should be inside source folder not
        # at folder where sphinx is run.

        if dotoctree:
            toctreestr = ':toctree: '
            api_dir = os.path.join(app.srcdir, app.config.automodapi_toctreedirnm)
            if docname is None:
                doc_path = '.'
            else:
                doc_path = os.path.join(app.srcdir, docname)
            toctreestr += os.path.relpath(api_dir, os.path.dirname(doc_path))
        else:
            toctreestr = ''

        newstrs = [spl[0]]
        for grp in range(len(spl) // 3):
            modnm = spl[grp * 3 + 1]

            # find where this is in the document for warnings
            if docname is None:
                location = None
            else:
                location = (docname, spl[0].count('\n'))

            # initialize default options
            toskip = []
            inhdiag = maindocstr = top_head = True
            hds = '-^'
            allowedpkgnms = []
            allowothers = False

            # look for actual options
            unknownops = []
            inherited_members = None
            for opname, args in _automodapiargsrex.findall(spl[grp * 3 + 2]):
                if opname == 'skip':
                    toskip.append(args.strip())
                elif opname == 'no-inheritance-diagram':
                    inhdiag = False
                elif opname == 'no-main-docstr':
                    maindocstr = False
                elif opname == 'headings':
                    hds = args
                elif opname == 'no-heading':
                    top_head = False
                elif opname == 'allowed-package-names':
                    allowedpkgnms.append(args.strip())
                elif opname == 'inherited-members':
                    inherited_members = True
                elif opname == 'no-inherited-members':
                    inherited_members = False
                elif opname == 'include-all-objects':
                    allowothers = True
                else:
                    unknownops.append(opname)

            # join all the allowedpkgnms
            if len(allowedpkgnms) == 0:
                allowedpkgnms = ''
                onlylocals = True
            else:
                allowedpkgnms = ':allowed-package-names: ' + ','.join(allowedpkgnms)
                onlylocals = allowedpkgnms

            # get the two heading chars
            if len(hds) < 2:
                msg = 'Not enough headings (got {0}, need 2), using default -^'
                if warnings:
                    app.warn(msg.format(len(hds)), location)
                hds = '-^'
            h1, h2 = hds.lstrip()[:2]

            # tell sphinx that the remaining args are invalid.
            if len(unknownops) > 0 and app is not None:
                opsstrs = ','.join(unknownops)
                msg = 'Found additional options ' + opsstrs + ' in automodapi.'
                if warnings:
                    app.warn(msg, location)

            ispkg, hascls, hasfuncs, hasother = _mod_info(
                modnm, toskip, onlylocals=onlylocals)

            # add automodule directive only if no-main-docstr isn't present
            if maindocstr:
                automodline = '.. automodule:: {modname}'.format(modname=modnm)
            else:
                automodline = ''
            if top_head:
                newstrs.append(automod_templ_modheader.format(
                        modname=modnm,
                        modhds=h1 * len(modnm),
                        pkgormod='Package' if ispkg else 'Module',
                        pkgormodhds=h1 * (8 if ispkg else 7),
                        automoduleline=automodline))  # noqa
            else:
                newstrs.append(automod_templ_modheader.format(
                    modname='',
                    modhds='',
                    pkgormod='',
                    pkgormodhds='',
                    automoduleline=automodline))

            # construct the options for the class/function sections
            # start out indented at 4 spaces, but need to keep the indentation.
            clsfuncoptions = []
            if toctreestr:
                clsfuncoptions.append(toctreestr)
            if toskip:
                clsfuncoptions.append(':skip: ' + ','.join(toskip))
            if allowedpkgnms:
                clsfuncoptions.append(allowedpkgnms)
            if hascls:  # This makes no sense unless there are classes.
                if inherited_members is True:
                    clsfuncoptions.append(':inherited-members:')
                if inherited_members is False:
                    clsfuncoptions.append(':no-inherited-members:')
            clsfuncoptionstr = '\n    '.join(clsfuncoptions)

            if hasfuncs:
                newstrs.append(automod_templ_funcs.format(
                    modname=modnm,
                    funchds=h2 * 9,
                    clsfuncoptions=clsfuncoptionstr))

            if hascls:
                newstrs.append(automod_templ_classes.format(
                    modname=modnm,
                    clshds=h2 * 7,
                    clsfuncoptions=clsfuncoptionstr))

            if allowothers and hasother:
                newstrs.append(automod_templ_vars.format(
                    modname=modnm,
                    otherhds=h2 * 9,
                    clsfuncoptions=clsfuncoptionstr))

            if inhdiag and hascls:
                # add inheritance diagram if any classes are in the module
                if toskip:
                    clsskip = ':skip: ' + ','.join(toskip)
                else:
                    clsskip = ''
                diagram_entry = automod_templ_inh.format(
                    modname=modnm,
                    clsinhsechds=h2 * 25,
                    allowedpkgnms=allowedpkgnms,
                    skip=clsskip)
                diagram_entry = diagram_entry.replace('    \n', '')
                newstrs.append(diagram_entry)

            newstrs.append(spl[grp * 3 + 3])

        newsourcestr = ''.join(newstrs)

        if app.config.automodapi_writereprocessed:
            # sometimes they are unicode, sometimes not, depending on how
            # sphinx has processed things
            if isinstance(newsourcestr, text_type):
                ustr = newsourcestr
            else:
                ustr = newsourcestr.decode(app.config.source_encoding)

            if docname is None:
                with io.open(os.path.join(app.srcdir, 'unknown.automodapi'),
                             'a', encoding='utf8') as f:
                    f.write(u'\n**NEW DOC**\n\n')
                    f.write(ustr)
            else:
                env = app.builder.env
                # Determine the filename associated with this doc (specifically
                # the extension)
                filename = docname + os.path.splitext(env.doc2path(docname))[1]
                filename += '.automodapi'

                with io.open(os.path.join(app.srcdir, filename), 'w',
                             encoding='utf8') as f:
                    f.write(ustr)

        return newsourcestr
    else:
        return sourcestr


def _mod_info(modname, toskip=[], onlylocals=True):
    """
    Determines if a module is a module or a package and whether or not
    it has classes or functions.
    """

    hascls = hasfunc = hasother = False

    for localnm, fqnm, obj in zip(*find_mod_objs(modname, onlylocals=onlylocals)):
        if localnm not in toskip:
            hascls = hascls or inspect.isclass(obj)
            hasfunc = hasfunc or inspect.isroutine(obj)
            hasother = hasother or (not inspect.isclass(obj) and
                                    not inspect.isroutine(obj))
            if hascls and hasfunc and hasother:
                break

    # find_mod_objs has already imported modname
    # TODO: There is probably a cleaner way to do this, though this is pretty
    # reliable for all Python versions for most cases that we care about.
    pkg = sys.modules[modname]
    ispkg = (hasattr(pkg, '__file__') and isinstance(pkg.__file__, str) and
             os.path.split(pkg.__file__)[1].startswith('__init__.py'))

    return ispkg, hascls, hasfunc, hasother


def process_automodapi(app, docname, source):
    source[0] = automodapi_replace(source[0], app, True, docname)


def setup(app):

    app.setup_extension('sphinx.ext.autosummary')

    # Note: we use __name__ here instead of just writing the module name in
    #       case this extension is bundled into another package
    from . import automodsumm
    app.setup_extension(automodsumm.__name__)

    app.connect('source-read', process_automodapi)

    app.add_config_value('automodapi_toctreedirnm', 'api', True)
    app.add_config_value('automodapi_writereprocessed', False, True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
# from __future__ import unicode_literals

import pytest
from functools import partial
import numpy as np
from numpy.testing import assert_equal, assert_allclose
from astropy.tests.helper import assert_quantity_allclose
from astropy import units as apu
from astropy.units import Quantity
from ... import conversions as cnv
from ... import pathprof
from ...utils import check_astro_quantities
from astropy.utils.misc import NumpyRNGContext
from geographiclib.geodesic import Geodesic


TOL_KWARGS = {'atol': 1.e-4, 'rtol': 1.e-4}


class TestGeodesics:

    def setup(self):

        pass

    def teardown(self):

        pass

    def test_inverse_cython(self):

        # testing against geographic-lib

        with NumpyRNGContext(1):

            lon1 = np.random.uniform(0, 360, 50)
            lon2 = np.random.uniform(0, 360, 50)
            lat1 = np.random.uniform(-90, 90, 50)
            lat2 = np.random.uniform(-90, 90, 50)

        distance, bearing1, bearing2 = pathprof.geodesics.inverse_cython(
            np.radians(lon1), np.radians(lat1),
            np.radians(lon2), np.radians(lat2),
            )
        (
            distance_lowprec, bearing1_lowprec, bearing2_lowprec
            ) = pathprof.geodesics.inverse_cython(
            np.radians(lon1), np.radians(lat1),
            np.radians(lon2), np.radians(lat2),
            eps=1.e-8
            )

        distance_gglib = np.empty_like(lon1)
        bearing1_gglib = np.empty_like(lon1)
        bearing2_gglib = np.empty_like(lon1)

        for idx, (_lon1, _lat1, _lon2, _lat2) in enumerate(zip(
                lon1, lat1, lon2, lat2
                )):

            aux = Geodesic.WGS84.Inverse(_lat1, _lon1, _lat2, _lon2)
            distance_gglib[idx] = aux['s12']
            bearing1_gglib[idx] = aux['azi1']
            bearing2_gglib[idx] = aux['azi2']

        assert_quantity_allclose(
            distance,
            distance_gglib,
            # atol=1.e-10, rtol=1.e-4
            )

        assert_quantity_allclose(
            distance_lowprec,
            distance_gglib,
            atol=1.,
            )

        assert_quantity_allclose(
            np.degrees(bearing1),
            bearing1_gglib,
            # atol=1.e-10, rtol=1.e-4
            )

        assert_quantity_allclose(
            np.degrees(bearing1_lowprec),
            bearing1_gglib,
            atol=1.e-6,
            )

        assert_quantity_allclose(
            np.degrees(bearing2),
            bearing2_gglib,
            # atol=1.e-10, rtol=1.e-4
            )

        assert_quantity_allclose(
            np.degrees(bearing2_lowprec),
            bearing2_gglib,
            atol=1.e-6,
            )

    def test_direct_cython(self):

        # testing against geographic-lib

        with NumpyRNGContext(1):

            lon1 = np.random.uniform(0, 360, 50)
            lat1 = np.random.uniform(-90, 90, 50)
            bearing1 = np.random.uniform(-90, 90, 50)
            dist = np.random.uniform(1, 10.e6, 50)  # 10000 km max

        lon2, lat2, bearing2 = pathprof.cygeodesics.direct_cython(
            np.radians(lon1), np.radians(lat1),
            np.radians(bearing1), dist
            )
        (
            lon2_lowprec, lat2_lowprec, bearing2_lowprec
            ) = pathprof.cygeodesics.direct_cython(
            np.radians(lon1), np.radians(lat1),
            np.radians(bearing1), dist, eps=1.e-8
            )

        lon2_gglib = np.empty_like(lon1)
        lat2_gglib = np.empty_like(lon1)
        bearing2_gglib = np.empty_like(lon1)

        for idx, (_lon1, _lat1, _bearing1, _dist) in enumerate(zip(
                lon1, lat1, bearing1, dist
                )):

            line = Geodesic.WGS84.Line(_lat1, _lon1, _bearing1)
            pos = line.Position(_dist)
            lon2_gglib[idx] = pos['lon2']
            lat2_gglib[idx] = pos['lat2']
            bearing2_gglib[idx] = pos['azi2']

        assert_quantity_allclose(
            np.degrees(lon2),
            lon2_gglib,
            # atol=1.e-10, rtol=1.e-4
            )

        assert_quantity_allclose(
            np.degrees(lon2_lowprec),
            lon2_gglib,
            atol=1.e-6,
            )

        assert_quantity_allclose(
            np.degrees(lat2),
            lat2_gglib,
            # atol=1.e-10, rtol=1.e-4
            )

        assert_quantity_allclose(
            np.degrees(lat2_lowprec),
            lat2_gglib,
            atol=1.e-6,
            )

        assert_quantity_allclose(
            np.degrees(bearing2),
            bearing2_gglib,
            # atol=1.e-10, rtol=1.e-4
            )

        assert_quantity_allclose(
            np.degrees(bearing2_lowprec),
            bearing2_gglib,
            atol=1.e-6,
            )
